-- When - a calendar
-- Copyright (C) 2023 Duane Robertson

-- when.lua

-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.


-- local dh = require 'duane_helper'


local when = {}


local DAY = 24 * 60 * 60
local DAYS = {'sun', 'mon', 'tue', 'wed', 'thu', 'fri', 'sat'}
local DAYS_LOOK = {}
local FILE_NAMES = {
	'calendar',
	'sample-calendar',
	(os.getenv('HOME') or '') .. '/.when/calendar',
	(os.getenv('HOME') or '') .. '/computing/love/when/calendar',
}
local FUTURE = 90.0 -- 365.0 * 10.0

local MONTHS = {'jan', 'feb', 'mar', 'apr', 'may', 'jun',
				'jul', 'aug', 'sep', 'oct', 'nov', 'dec'}
local MONTH_LEN = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
local MONTHS_LOOK = {}

local OP_PREC = {
	['('] = 8, [')'] = 8,
	['%'] = 7,
	['-'] = 6,
	['<'] = 5, ['>'] = 5, ['<='] = 5, ['>='] = 5,
	['='] = 4, ['!='] = 4,
	['!'] = 3,
	['&'] = 2,
	['|'] = 1,
	}


for k, v in pairs(DAYS) do
	DAYS_LOOK[v] = k
end

for k, v in pairs(MONTHS) do
	MONTHS_LOOK[v] = k
end


local function main()
	local items = when:parseCalendar()
	local date = tonumber(os.time())
	local q = when:calendar(items, date)
	when:print_cal(q)
end


local function table_copy(t, seen)
    local n = {}
    seen = seen or {}
    seen[t] = n

    for k, v in pairs(t) do
        n[(type(k) == "table" and (seen[k] or table_copy(k, seen))) or k] =
        (type(v) == "table" and (seen[v] or table_copy(v, seen))) or v
    end

    return n
end


function when.calendar(self, items, date, future)
	local outq = {}
	future = future or FUTURE

	for dt = date - DAY, date + DAY * future, DAY do
		local tab = os.date('*t', dt)
		tab.hour = 12
		self:getJulian(tab)
		-- print(dt)
		-- print(dh.dump(tab))
		-- print(os.date('%d %b %Y', dt))
		-- print(dh.dump(tab))

		for _, item in pairs(items) do
			local txt
			-- print(dh.dump(item))

			if item.math then
				txt = self:matchRPN(item, tab)
			elseif item.year then
				-- print(dh.dump(item))
				txt = self:matchDMY(item, tab)
			end

			if txt then
				local outi = {
					dt = dt,
					text = txt,
				}

				-- print(item.math)
				local sday = os.date('%a', dt)
				if dt - date < 0 then
					sday = 'Yesterday'
				elseif dt - date < DAY then
					sday = 'Today'
				elseif dt - date < DAY * 2 then
					sday = 'Tomorrow'
				end
				sday = (sday .. (' '):rep(10)):sub(1, 10)
				outi.date = sday .. os.date('%Y %b %e', dt)

				outq[#outq+1] = outi
			end
		end

		-- break
	end

	return outq
end


function when.getDaysInMonth(self, date)
	local ds = MONTH_LEN[date.month]
	if date.month == 2 and date.year % 4 == 0 then
		ds = ds + 1
	end
	return ds
end


function when.getJulian(self, datet)
	local p1 = math.modf((datet.month - 14) / 12)
	local p2 = datet.year + 4800 + p1
	local jul = math.modf((1461 * p2) / 4)
			+ math.modf((367 * (datet.month - 2 - 12 * p1)) / 12)
			- math.modf((3 * math.modf((p2 + 100) / 100)) / 4)
			+ datet.day - 32075 - 2400000
	jul = math.modf(jul)

	----------------------------------------------
	-- When uses modified julian, so this is wrong.
	----------------------------------------------
	local jn = jul + (datet.hour - 12) / 24
			+ datet.min / 1440 + datet.sec / 86400
	jn = math.floor(jn)
	----------------------------------------------

	datet.jn = jn
	datet.julian = jul
end

if false then
	local tests = {
		{ month = 4, day = 29, year = 2023, julian = 60064 },
		{ month = 5, day = 1, year = 2023, julian = 60066 },
		{ month = 5, day = 1, year = 2024, julian = 60432 },
		{ month = 12, day = 1, year = 2024, julian = 60646 },
		{ month = 11, day = 30, year = 2024, julian = 60645 },
	}
	for _, t in pairs(tests) do
		local t1 = {
			month = t.month,
			day = t.day,
			year = t.year,
			hour = 12,
			min = 1,
			sec = 1,
		}
		when.getJulian({}, t1)
		assert(t1.julian == t.julian, 'bad day')
	end
end


function when.getWeek(self, date, reverse)
	local a = math.ceil(date.day / 7)

	if reverse then
		local ds = self:getDaysInMonth(date)
		-- print(date.month, self:getDaysInMonth(date))
		a = tonumber(math.ceil((ds - date.day + 1) / 7))
	end

	-- print(date.month, date.day, reverse, a)
	return a
end


function when.makeRPN(self, str)
	-- Change the infix expression to reverse
	--  polish using the shunting yard algorithm.

	local toks = self:tokenize(str)

	-- print(table.concat(toks, ','))

	local outq = {}
	local opq = {}

	while #toks > 0 do
		local t = table.remove(toks, 1)

		if t == '(' then
			opq[#opq+1] = t

		elseif t == ')' then
			while #opq > 0 and opq[#opq] ~= '(' do
				assert(#opq ~= 0, 'opq error 1')
				outq[#outq+1] = table.remove(opq)
			end
			assert(opq[#opq] == '(', 'opq error 2')
			table.remove(opq)

		elseif OP_PREC[t] then
			while #opq > 0 and opq[#opq] ~= '(' and OP_PREC[opq[#opq]] >= OP_PREC[t] do
				outq[#outq+1] = table.remove(opq)
			end
			opq[#opq+1] = t
		else
			if t:match('^%d+$') then
				outq[#outq+1] = tonumber(t)
			elseif MONTHS_LOOK[t:lower()] then
				outq[#outq+1] = MONTHS_LOOK[t:lower()]
			elseif DAYS_LOOK[t:lower()] then
				outq[#outq+1] = DAYS_LOOK[t:lower()]
			else
				outq[#outq+1] = t
			end
		end
	end

	while #opq > 0 do
		assert(opq[#opq] ~= '(', 'parenthesis error')
		outq[#outq+1] = table.remove(opq)
	end

	-- print(table.concat(outq, ','))

	return outq
end


function when.matchDMY(self, item, date)
	-- print(dh.dump(item))
	if item.year ~= date.year
	and (not item.year_plus or date.year < item.year) then
		return
	end

	if item.month ~= date.month
	and (not item.month_plus or date.month < item.month) then
		return
	end

	if item.day ~= date.day
	and (not item.day_plus or date.day < item.day) then
		return
	end

	-- print(dh.dump(item))
	local txt
	if item.year_plus and item.year then
		local num = date.year - item.year

		local suff = 'th'
		if num % 10 == 1 and num % 100 ~= 11 then
			suff = 'st'
		elseif num % 10 == 2 and num % 100 ~= 12 then
			suff = 'nd'
		elseif num % 10 == 3 and num % 100 ~= 13 then
			suff = 'rd'
		end

		local num_str = tostring(num) .. suff
		txt = item.text:gsub('\\a', num_str)
	end

	return txt or item.text
end


function when.matchRPN(self, item, date)
	local stack = {}
	local q = table_copy(item.rpn)
	-- print(item.text, item.math)
	-- print(dh.dump(q))

	while #q > 0 do
		local t = table.remove(q, 1)
		if OP_PREC[t] then
			local a, o1, o2

			if t == '!' then  -- only unary operator
				o1 = table.remove(stack, 1)
			else
				o2 = table.remove(stack, 1)
				o1 = table.remove(stack, 1)
			end

			if t == '!' then
				a = (not o1) or (o1 == 0)
			elseif t == '%' then
				a = o1 % o2
			elseif t == '-' then
				a = o1 - o2
			elseif t == '=' then
				a = (o1 == o2)
			elseif t == '!=' then
				a = (o1 ~= o2)
			elseif t == '&' then
				a = (o1 and o2) and true or false
			elseif t == '|' then
				a = (o1 or o2) and true or false
			elseif t == '<' then
				a = (o1 < o2)
			elseif t == '>' then
				a = (o1 > o2)
			elseif t == '<=' then
				a = (o1 <= o2)
			elseif t == '>=' then
				a = (o1 >= o2)
			else
				assert(false, 'operator error')
			end

			if a ~= nil then
				table.insert(stack, 1, a)
			end
		else
			table.insert(stack, 1, self:replaceVariables(t, date))
		end
	end

	assert(#stack == 1, 'stack error')
	if table.remove(stack, 1) then
		return item.text
	end
end


function when.parseCalendar(self)
	local ct = 0
	local items = {}
	local file_name
	local f, ferr

	for _, fn in ipairs(FILE_NAMES) do
		f, ferr = io.open(fn)
		if f then
			file_name = fn
			f:close()
			break
		end
	end

	if not file_name then
		assert(false, ferr)
		return
	end

	for lin in io.lines(file_name) do
		ct = ct + 1
		local item = self:parseEntry(lin, ct)
		if item then
			items[#items+1] = item
		end
	end

	return items
end


function when.parseEntry(self, lin, num)
	local matSimpleDate = '^%s*(%d-%*?)%s+(%a-%*?)%s+(%d-%*?)%s*$'
	local lineMatch = '^%s*([^%c,]-)%s*,%s*([%w%p%s]-)%s*$'

	if lin:find('^#') then
		-- skip the comment
	elseif lin:find(',') then
		local dat, txt = lin:match(lineMatch)
		local year, month, day = dat:lower():match(matSimpleDate)
		local item = {}

		if year and month and day then
			item.year = tonumber(year:match('[^%*]+'))
			item.year_plus = year:match('%*')

			item.month = MONTHS_LOOK[month]
			if not item.month then
				item.month = tonumber(month:match('[^%*]+'))
			end

			item.month_plus = month:match('%*')
			item.day = tonumber(day:match('[^%*]+'))
			item.day_plus = day:match('%*')
			item.text = txt
			-- print(dh.dump(item))
		else
			item.math = dat
			item.rpn = self:makeRPN(dat)
			item.text = txt
			-- print(dh.dump(item))
		end

		return item
	else
		num = num or '?'
		print('** error in line ' .. num .. ': ' .. lin)
	end
end


function when.print_cal(self, outq)
	local already = {}
	local txt_green = '\027[1;32m'
	local txt_norm = '\027[0m'

	print(os.date('%a %Y %b %e %H:%M:%S\n'))

	for _, v in ipairs(outq) do
		if not already[v.text] then
			local dat = v.date
			if v.date:match('Today') or v.date:match('Tomorrow') then
				print(txt_green .. v.date .. ' ' .. v.text .. txt_norm)
			else
				print(dat .. ' ' .. v.text)
			end
			already[v.text] = true
		end
	end
end


function when.replaceVariables(self, t, date)
	if type(t) ~= 'string' then
		-- print('not string ' .. tostring(t))
		return t
	end

	if t == 'j' then
		t = date.julian
	elseif t == 'm' then
		t = date.month
	elseif t == 'd' then
		t = date.day
	elseif t == 'w' then
		t = date.wday
	elseif t == 'y' then
		t = date.year
	elseif t == 'a' then
		t = self:getWeek(date, false)
		-- print(date.month, date.day, t, 'a resolves to ' .. t)
	elseif t == 'b' then
		t = self:getWeek(date, true)
	elseif t == 'e' then
		------------------------
		-- Add Easter... someday.
		------------------------
		t = 999999
		------------------------
	end

	return t
end


function when.tokenize(self, str)
	local tok = {}
	local p = 1

	while p <= #str do
		local n1 = str:sub(p, p)
		local n2 = str:sub(p, p+1)

		local alph = str:sub(p):match('^%w+')
		if n1 == ' ' then
			p = p + 1
		elseif alph then
			tok[#tok+1] = alph
			p = p + #alph
		elseif OP_PREC[n2] then
			tok[#tok+1] = n2
			p = p + 2
		elseif OP_PREC[n1] then
			tok[#tok+1] = n1
			p = p + 1
		else
			print('** token error', '['..n1..']', '['..n2..']')
			p = p + 1
		end
	end

	return tok
end


if arg and arg[0] and arg[0]:match('when%.lua$') then
	main()
end

return when
